import asyncio

from yarl import URL
from aiohttp.web_runner import BaseSite, AppRunner, GracefulExit


class AGISite(BaseSite):
    __slots__ = ('_host', '_port', '_reuse_address', '_reuse_port')

    def __init__(self, runner, host=None, port=None, *,
                 shutdown_timeout=60.0, ssl_context=None,
                 backlog=128, reuse_address=None,
                 reuse_port=None):
        super().__init__(runner, shutdown_timeout=shutdown_timeout,
                         ssl_context=ssl_context, backlog=backlog)
        if host is None:
            host = "0.0.0.0"
        self._host = host

        if port is None:
            port = 8080
        self._port = port

        self._reuse_address = reuse_address
        self._reuse_port = reuse_port

    @property
    def name(self):
        return str(URL.build(scheme='agi', host=self._host, port=self._port))

    async def start(self):
        await super().start()
        loop = asyncio.get_event_loop()
        self._server = await loop.create_server(
            self._runner.server, self._host, self._port,
            ssl=self._ssl_context, backlog=self._backlog,
            reuse_address=self._reuse_address,
            reuse_port=self._reuse_port)


def run_app(app, *, host=None, port=None,
            shutdown_timeout=60.0,
            print=print, backlog=128, handle_signals=True,
            reuse_address=None, reuse_port=None):

    loop = asyncio.get_event_loop()

    if asyncio.iscoroutine(app):
        app = loop.run_until_complete(app)

    runner = AppRunner(app, handle_signals=handle_signals)
    loop.run_until_complete(runner.setup())

    try:
        loop.run_until_complete(get_site(
            runner,
            host=host, port=port,
            shutdown_timeout=shutdown_timeout, backlog=backlog,
            reuse_address=reuse_address, reuse_port=reuse_port,
        ))
        try:
            if print:  # pragma: no branch
                names = sorted(str(s.name) for s in runner.sites)
                print("======== Running on {} ========\n"
                      "(Press CTRL+C to quit)".format(', '.join(names)))
            loop.run_forever()
        except (GracefulExit, KeyboardInterrupt):  # pragma: no cover
            pass
    finally:
        loop.run_until_complete(runner.cleanup())
    if hasattr(loop, 'shutdown_asyncgens'):
        loop.run_until_complete(loop.shutdown_asyncgens())
    loop.close()


async def get_site(runner, *,
                   host=None, port=None,
                   shutdown_timeout=60.0, backlog=128,
                   reuse_address=None, reuse_port=None):

    site = AGISite(runner, host, port,
                   shutdown_timeout=shutdown_timeout,
                   backlog=backlog,
                   reuse_address=reuse_address,
                   reuse_port=reuse_port)

    await site.start()



from enum import Enum


class AGIMethod(str, Enum):
    ANY = '*'
    LOCAL = 'LOCAL'
    SIP = 'SIP'
    DAHDI = 'DAHDI'
    IAX2 = 'IAX2'
    MOTIF = 'MOTIF'
    MISDN = 'MISDN'
    USTM = 'USTM'
    SKINNY = 'SKINNY'
    H323 = 'H323'
    OOH323 = 'OOH323'
    GTALK = 'GTALK'
    JINGLE = 'JINGLE'

    @classmethod
    def all(cls):
        return {
            cls.LOCAL, cls.SIP, cls.DAHDI, cls.IAX2,
            cls.MOTIF, cls.MISDN, cls.USTM, cls.SKINNY,
            cls.H323, cls.OOH323, cls.GTALK, cls.JINGLE,
        }



import asyncio


class AGIClientError(Exception):
    """Base class for client connection errors."""


class AGIInvalidURL(AGIClientError, ValueError):
    """Invalid URL.

    URL used for fetching is malformed, e.g. it doesn't contains host
    part."""

    # Derive from ValueError for backward compatibility

    def __init__(self, url):
        super().__init__(url)

    @property
    def url(self):
        return self.args[0]

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.url)


class AGIInvalidMethod(AGIClientError, ValueError):
    pass


class AGIClientConnectionError(AGIClientError):
    """Base class for client socket errors."""


class AGIServerConnectionError(AGIClientConnectionError):
    """Server connection errors."""


class AGIServerTimeoutError(AGIServerConnectionError, asyncio.TimeoutError):
    """Server timeout error."""



from aiohttp.abc import AbstractStreamWriter

from aioagi.commands import AGICommand


class AGIPayloadWriter(AbstractStreamWriter):

    def __init__(self, protocol, loop):
        self._protocol = protocol
        self._transport = protocol.transport

        self.loop = loop
        self._eof = False

    @property
    def transport(self):
        return self._transport

    @property
    def protocol(self):
        return self._protocol

    async def write(self, chunk):
        """Writes chunk of data to a stream.

        write_eof() indicates end of stream.
        writer can't be used after write_eof() method being called.
        write() return drain future.
        """
        chunk = chunk.encode('utf-8')
        if chunk:
            if self._transport is None or self._transport.is_closing():
                raise ConnectionResetError('Cannot write to closing transport')
            self._transport.write(chunk)

    async def write_headers(self, headers, SEP=': ', END='\n'):
        """Write request headers."""
        # status + headers
        headers = ''.join(
            [k + SEP + v + END for k, v in headers.items()]
        )
        headers = headers + '\n'

        await self.write(headers)

    async def write_eof(self, chunk=b''):
        if self._eof:
            return

        if not chunk:
            chunk = AGICommand.HANGUP.value.encode('utf-8')
            chunk = chunk + b'\n'

        self.write(chunk)
        await self.drain(True)

        self._eof = True
        self._transport = None

    async def drain(self, last=False):
        """Flush the write buffer.

        The intended use is to write

          await w.write(data)
          await w.drain()
        """
        if self._protocol.transport is not None:
            await self._protocol._drain_helper()



"""
This module contains functions and classes to implment AGI scripts in python.
"""
import asyncio

from aiohttp import EofStream

from aioagi.commands import AGICommand
from aioagi.exceptions import (
    AGIResultHangup, AGIConnectHangup,
    AGIUsageError, AGIInvalidCommand,
    AGIAppError, AGICommandNotPermittedError
)
from aioagi.parser import AGIMessage, AGICode
from aioagi.request import AGIRequest

DEFAULT_TIMEOUT = 2000  # 2sec timeout used as default for functions that take timeouts
DEFAULT_RECORD = 20000  # 20sec record time


class AGI:
    """
    This class encapsulates communication between Asterisk an a python script.
    It handles encoding commands to Asterisk and parsing responses from
    Asterisk.
    """

    def __init__(self, request: AGIRequest):
        """
        request.headers = {
            'agi_network': 'yes',
            'agi_network_script': 'agi/',
            'agi_request': 'agi://10.99.99.55:8080/agi/',
            'agi_channel': 'SIP/95000-00000005',
            'agi_language': 'ru',
            'agi_type': 'SIP',
            'agi_uniqueid': '1502266948.5',
            'agi_version': '11.14.2',
            'agi_callerid': '95000',
            'agi_calleridname': 'test',
            'agi_callingpres': '0',
            'agi_callingani2': '0',
            'agi_callington': '0',
            'agi_callingtns': '0',
            'agi_dnid': '770',
            'agi_rdnis': 'unknown',
            'agi_context': 'from-internal',
            'agi_extension': '770',
            'agi_priority': '2',
            'agi_enhanced': '0.0',
            'agi_accountcode': '',
            'agi_threadid': '139689736754944',
        }
        """
        self.request = request
        self.headers = request.headers

    def _quote(self, string):
        return '"{}"'.format(str(string))

    def _process_digit_list(self, digits):
        if isinstance(digits, (list, tuple)):
            digits = ''.join(map(str, digits))
        return self._quote(digits)

    async def run_command(self, command: AGICommand, *args)-> AGIMessage:
        """
        Run a command to Asterisk

        If any command returned `-1` - raise AGIAppError.
        This means that chanel return error or hangup.
        """
        command = command.value
        command = '{} {}'.format(command, ' '.join(map(str, args)))
        command = command.strip()
        if command[-1] != '\n':
            command += '\n'

        await self.request.writer.write(command)
        await self.request.writer.drain()
        try:
            message = await self.request.reader.read()
            if message.status == AGICode.TRYING:
                message = await self.request.reader.read()

        except EofStream:
            raise AGIResultHangup()
        except asyncio.CancelledError:
            raise AGIConnectHangup()

        if message.status == AGICode.MEMORY_ALLOCATION_FAILURE:
            raise AGIInvalidCommand(error=message.info)
        elif message.status == AGICode.INVALID_COMMAND:
            raise AGIInvalidCommand(error=message.info)
        elif message.status == AGICode.COMMAND_NOT_PERMITTED:
            raise AGICommandNotPermittedError(error=message.info)
        elif message.status == AGICode.INVALID_COMMAND_SYNTAX:
            raise AGIUsageError(error=message.info)
        elif message.result == '-1':
            raise AGIAppError(message='Error executing application, or hangup.'.encode())

        return message

    async def answer(self):
        """
        Answer channel

        Answers channel if not already in answer state.
        Returns -1 on channel failure, or 0 if successful.
        """
        return await self.run_command(AGICommand.ANSWER)

    async def channel_status(self, channel=''):
        """
        Returns status of the connected channel.

        Returns the status of the specified channelname.
        If no channel name is given then returns the status of the current channel.

        Return values:
        0 - Channel is down and available.
        1 - Channel is down, but reserved.
        2 - Channel is off hook.
        3 - Digits (or equivalent) have been dialed.
        4 - Line is ringing.
        5 - Remote end is ringing.
        6 - Line is up.
        7 - Line is busy.
        """
        message = await self.run_command(AGICommand.CHANNEL_STATUS, channel)
        return message.result_to_int()

    async def control_stream_file(self, filename, escape_digits='', skipms=3000, fwd='', rew='', pause=''):
        """
        Sends audio file on channel and allows the listener to control the stream.

        Use double quotes for the digits if you wish none to be permitted.
        Returns 0 if playback completes without a digit being pressed,
        or the ASCII numerical value of the digit if one was pressed,
        or -1 on error or if the channel was disconnected.

        Send the given file, allowing playback to be interrupted by the given
        digits, if any.  escape_digits is a string '12345' or a list  of
        ints [1,2,3,4,5] or strings ['1','2','3'] or mixed [1,'2',3,'4']
        If sample offset is provided then the audio will seek to sample
        offset before play starts.  Returns  digit if one was pressed.
        Throws AGIAppError if the channel was disconnected.  Remember, the file
        extension must not be included in the filename.
        """
        message = await self.run_command(
            AGICommand.CONTROL_STREAM_FILE,
            self._quote(filename),
            self._process_digit_list(escape_digits),
            self._quote(skipms),
            self._quote(fwd),
            self._quote(rew),
            self._quote(pause)
        )
        return message.result_to_char()

    async def database_del(self, family, key):
        """
        Removes database key/value.

        Deletes an entry in the Asterisk database for a
        given family and key.
        Returns 1 if successful, 0 otherwise.
        """
        message = await self.run_command(
            AGICommand.DATABASE_DEL,
            self._quote(family),
            self._quote(key)
        )
        return message.result_to_bool()

    async def database_deltree(self, family, key=''):
        """
        Removes database keytree/value.

        Deletes a family or specific keytree with in a family in the Asterisk database.
        Returns 1 if successful, 0 otherwise.
        """
        message = await self.run_command(
            AGICommand.DATABASE_DELTREE,
            self._quote(family),
            self._quote(key)
        )
        return message.result_to_bool()

    async def database_get(self, family, key):
        """
        Gets database value.

        Retrieves an entry in the Asterisk database for a given family and key.
        Returns 0 if <key> is not set.  Returns 1 if <key>
        is set and returns the variable in parenthesis.
        Example return code: `200 result=1 (testvariable)`.
        """
        message = await self.run_command(
            AGICommand.DATABASE_GET,
            self._quote(family),
            self._quote(key)
        )
        return message.result_to_bool()

    async def database_put(self, family, key, value):
        """
        Adds/updates database value

        Adds or updates an entry in the Asterisk database for a
        given family, key, and value.
        Returns 1 if successful, 0 otherwise.
        """
        message = await self.run_command(
            AGICommand.DATABASE_PUT,
            self._quote(family),
            self._quote(key),
            self._quote(value)
        )
        return message.result_to_bool()

    async def exec_app(self, application, options=''):
        """
        Executes a given Application.

        Executes <application> with given <options>.
        Returns whatever the application returns, or -2 on failure to find
        application
        """
        message = await self.run_command(AGICommand.EXEC, application, self._quote(options))
        if message.result == '-2':
            raise AGIAppError(message='Unable to find application: %s' % application)
        return message

    async def get_data(self, filename, timeout=DEFAULT_TIMEOUT, max_digits=255):
        """
        Prompts for DTMF on a channel.

        Stream the given file, and receive DTMF data.
        Returns the digits received from the channel at the other end.
        Example: `200 result=132456`.
        """
        return await self.run_command(AGICommand.GET_DATA, filename, timeout, max_digits)

    async def get_full_variable(self, name, channel=None):
        """
        Evaluates a channel expression.

        Returns 0 if variablename is not set or channel does not exist.
        Returns 1 if variablename is set and returns the variable in parenthesis.
        Understands complex variable names and builtin variables, unlike GET VARIABLE.

        Example return code: `200 result=1 (testvariable)`
        """
        args = [AGICommand.GET_FULL_VARIABLE, self._quote(name)]
        if channel:
            args.append(self._quote(channel))

        message = await self.run_command(*args)
        return message.result_to_bool()

    async def get_option(self, filename, escape_digits='', timeout=0):
        """
        Stream file, prompt for DTMF, with timeout.

        Behaves similar to STREAM FILE but used with a timeout option.

        Send the given file, allowing playback to be interrupted by the given
        digits, if any.  escape_digits is a string '12345' or a list  of
        ints [1,2,3,4,5] or strings ['1','2','3'] or mixed [1,'2',3,'4']
        Returns  digit if one was pressed.
        Remember, the file extension must not be included in the filename.
        """
        escape_digits = self._process_digit_list(escape_digits)
        args = [AGICommand.GET_OPTION, filename, escape_digits]
        if timeout:
            args.append(timeout)

        message = await self.run_command(*args)
        return message.result_to_char()

    async def get_variable(self, name):
        """
        Get a channel variable.

        Returns 0 if variablename is not set.
        Returns 1 if variablename is set and returns the variable in parentheses.
        Example return code: `200 result=1 (testvariable)`.

        This function returns the value of the indicated channel variable.
        """
        message = await self.run_command(
            AGICommand.GET_VARIABLE,
            self._quote(name)
        )

        return message.result_to_bool()

    async def gosub(self, context, extension, priority, *args):
        """
        Cause the channel to execute the specified dialplan subroutine.

        Cause the channel to execute the specified dialplan subroutine,
        returning to the dialplan with execution of a Return().
        """
        message = await self.run_command(
            AGICommand.GOSUB,
            context, extension, priority,
            '{}'.format(','.join(map(str, args)))
        )
        return message

    async def goto_on_exit(self, context='', extension='', priority=''):
        """
        Set endpoint in dialplan on exit.
        Group execute `set_context`, `set_extension` and `set_priority`.
        """
        await self.set_context(context or self.headers['agi_context'])
        await self.set_extension(extension or self.headers['agi_extension'])
        await self.set_priority(priority or self.headers['agi_priority'])

    async def hangup(self, channel=''):
        """
        Hangup a channel.

        Hangs up the specified channel.
        If no channel name is given, hangs up the current channel.
        """
        return await self.run_command(AGICommand.HANGUP, channel)

    async def noop(self):
        """
        Does nothing.
        """
        return await self.run_command(AGICommand.NOOP)

    async def receive_char(self, timeout=DEFAULT_TIMEOUT):
        """
        Receives one character from channels supporting it.

        Returns the decimal value of the character if one is received,
        or 0 if the channel does not support text reception.
        Returns -1 only on error/hangup.

        Receives a character of text on a channel. Specify timeout to be the
        maximum time to wait for input in milliseconds, or 0 for infinite.
        Most channels do not support the reception of text.
        """
        message = await self.run_command(AGICommand.RECEIVE_CHAR, timeout)
        return message.result_to_char()

    async def receive_text(self, timeout=DEFAULT_TIMEOUT):
        """
        Receives text from channels supporting it.

        Receives a string of text on a channel.
        Most channels do not support the reception of text.
        Returns -1 for failure or 1 for success, and the string in parenthesis.
        """
        message = await self.run_command(AGICommand.RECEIVE_TEXT, timeout)
        return message

    async def record_file(self,
                          filename, audio_format='gsm',
                          escape_digits='#', timeout=DEFAULT_RECORD, offset=0,
                          beep='beep', silence: int=None):
        """
        Records to a given file.

        Record to a file until a given dtmf digit in the sequence is received
        The format will specify what kind of file will be recorded.  The timeout
        is the maximum record time in milliseconds, or -1 for no timeout. Offset
        samples is optional, and if provided will seek to the offset without
        exceeding the end of the file

        :param filename: The destination filename of the recorded audio.
        :param audio_format: The audio format in which to save the resulting file.
        :param escape_digits: The DTMF digits that will terminate the recording process.
        :param timeout: The maximum recording time in milliseconds. Set to -1 for no limit.
        :param offset: Causes the recording to first seek to the specified offset before recording begins.
        :param beep: Causes Asterisk to play a beep as recording begins. This argument can take any value.
        :param silence: The number of seconds of silence that are permitted before the recording is terminated,
                        regardless of the escape_digits or timeout arguments.
                        If specified, this parameter must be preceded by s=.
        """
        escape_digits = self._process_digit_list(escape_digits)
        args = [
            AGICommand.RECORD_FILE, self._quote(filename), audio_format,
            escape_digits, timeout, offset, beep,
        ]
        if silence:
            args.append('s={}'.format(silence))

        return await self.run_command(*args)

    async def say_alpha(self, characters, escape_digits=''):
        """
        Says a given character string.

        Say a given character string, returning early if any of the given DTMF
        digits are received on the channel.
        Throws AGIAppError on channel failure
        """
        characters = self._process_digit_list(characters)
        escape_digits = self._process_digit_list(escape_digits)
        message = await self.run_command(AGICommand.SAY_ALPHA, characters, escape_digits)
        return message.result_to_char()

    async def say_date(self, seconds, escape_digits=''):
        """
        Says a given date.

        Say a given date, returning early if any of the given DTMF digits are
        received on the channel.
        Returns 0 if playback completes without a digit being pressed,
        or the ASCII numerical value of the digit if one was pressed or -1 on error/hangup.

        The date should be in seconds since the UNIX Epoch (Jan 1, 1970 00:00:00)

        :param seconds: Is number of seconds elapsed since 00:00:00 on January 1, 1970.
                        Coordinated Universal Time (UTC).
        """
        escape_digits = self._process_digit_list(escape_digits)
        message = await self.run_command(AGICommand.SAY_DATE, seconds, escape_digits)
        return message.result_to_char()

    async def say_datetime(self, seconds, escape_digits='', format='', zone=''):
        """
        Says a given time as specified by the format given.

        Say a given time, returning early if any of the given DTMF digits are received on the channel.
        Returns 0 if playback completes without a digit being pressed,
        or the ASCII numerical value of the digit if one was pressed or -1 on error/hangup.

        :param seconds: Is number of seconds elapsed since 00:00:00 on January 1, 1970, Coordinated Universal Time (UTC)
        :param escape_digits: экранированные цифры
        :param format: Is the format the time should be said in. See voicemail.conf (defaults to ABdY 'digits/at' IMp).
        :param zone: Acceptable values can be found in /usr/share/zoneinfo Defaults to machine default.
        """
        escape_digits = self._process_digit_list(escape_digits)
        if format:
            format = self._quote(format)

        message = await self.run_command(
            AGICommand.SAY_DATETIME,
            seconds,
            escape_digits,
            format,
            zone
        )
        return message.result_to_char()

    async def say_digits(self, digits, escape_digits=''):
        """
        Says a given digit string.

        Say a given digit string, returning early if any of the given DTMF digits are received on the channel.
        Returns 0 if playback completes without a digit being pressed,
        or the ASCII numerical value of the digit if one was pressed or -1 on error/hangup.
        """
        digits = self._process_digit_list(digits)
        escape_digits = self._process_digit_list(escape_digits)
        message = await self.run_command(AGICommand.SAY_DIGITS, digits, escape_digits)
        return message.result_to_char()

    async def say_number(self, number, escape_digits=''):
        """
        Says a given number.

        Say a given number, returning early if any of the given DTMF digits are received on the channel.
        Returns 0 if playback completes without a digit being pressed,
        or the ASCII numerical value of the digit if one was pressed or -1 on error/hangup.
        """
        number = self._process_digit_list(number)
        escape_digits = self._process_digit_list(escape_digits)
        message = await self.run_command(AGICommand.SAY_NUMBER, number, escape_digits)
        return message.result_as_char()

    async def say_phonetic(self, characters, escape_digits=''):
        """
        Says a given character string with phonetics.

        Say a given character string with phonetics,
        returning early if any of the given DTMF digits are received on the channel.
        Returns 0 if playback completes without a digit pressed,
        the ASCII numerical value of the digit if one was pressed, or -1 on error/hangup.
        """
        characters = self._process_digit_list(characters)
        escape_digits = self._process_digit_list(escape_digits)
        message = await self.run_command(
            AGICommand.SAY_PHONETIC, characters, escape_digits
        )
        return message.result_to_char()

    async def say_time(self, seconds, escape_digits=''):
        """
        Says a given time.

        Say a given time, returning early if any of the given DTMF digits are received on the channel.
        Returns 0 if playback completes without a digit being pressed,
        or the ASCII numerical value of the digit if one was pressed or -1 on error/hangup.

        :param time: Is number of seconds elapsed since 00:00:00 on January 1, 1970. Coordinated Universal Time (UTC).
        :param escape_digits: экранированные числа
        """
        escape_digits = self._process_digit_list(escape_digits)
        message = await self.run_command(AGICommand.SAY_TIME, seconds, escape_digits)
        return message.result_to_char()

    async def send_image(self, filename):
        """
        Sends images to channels supporting it.

        Sends the given image on a channel.
        Most channels do not support the transmission of images.
        Returns 0 if image is sent, or if the channel does not support image transmission.
        Returns -1 only on error/hangup. Image names should not include extensions.
        """
        return await self.run_command(AGICommand.SEND_IMAGE, filename)

    async def send_text(self, text=''):
        """
        Sends text to channels supporting it.

        Sends the given text on a channel.
        Most channels do not support the transmission of text.
        Returns 0 if text is sent, or if the channel does not support text transmission.
        Returns -1 only on error/hangup.
        """
        return await self.run_command(AGICommand.SEND_TEXT, self._quote(text))

    async def set_autohangup(self, secs):
        """
        Autohangup channel in some time.

        Cause the channel to automatically hangup at <secs> seconds in the
        future.  Of course it can be hungup before then as well. Setting to
        0 will cause the autohangup feature to be disabled on this channel.
        """
        await self.run_command(AGICommand.SET_AUTOHANGUP, secs)

    async def set_callerid(self, number):
        """
        Sets callerid for the current channel.

        Changes the callerid of the current channel.
        """
        await self.run_command(AGICommand.SET_CALLERID, number)

    async def set_context(self, context):
        """
        Sets channel context.

        Sets the context for continuation upon exiting the application.
        No error appears to be produced.  Does not set exten or priority
        Use at your own risk.  Ensure that you specify a valid context.
        """
        await self.run_command(AGICommand.SET_CONTEXT, context)

    async def set_music(self, toggle, music='default'):
        """
        Enable/Disable Music on hold generator
        Enables/Disables the music on hold generator.
        If class is not specified, then the default music on hold class will be used.
        This generator will be stopped automatically when playing a file.
        Always returns 0.

        Usage: SET MUSIC <on|off> <class>

        :param music: music class
        :param toggle: on|off
        """
        return await self.run_command(AGICommand.SET_MUSIC, toggle, music)

    async def set_extension(self, extension):
        """
        Changes channel extension.

        Sets the extension for continuation upon exiting the application.
        No error appears to be produced.  Does not set context or priority
        Use at your own risk.  Ensure that you specify a valid extension.
        """
        await self.run_command(AGICommand.SET_EXTENSION, extension)

    async def set_priority(self, priority):
        """
        Set channel dialplan priority.

        Sets the priority for continuation upon exiting the application.
        No error appears to be produced.  Does not set exten or context
        Use at your own risk.  Ensure that you specify a valid priority.
        """
        await self.run_command(AGICommand.SET_PRIORITY, priority)

    async def set_variable(self, name, value):
        """
        Set a channel variable.
        Sets a variable to the current channel.
        """
        await self.run_command(AGICommand.SET_VARIABLE, self._quote(name), self._quote(value))

    async def stream_file(self, filename, escape_digits='', sample_offset=0):
        """
        Sends audio file on channel.

        Send the given file, allowing playback to be interrupted by the given digits, if any.
        Returns 0 if playback completes without a digit being pressed,
        or the ASCII numerical value of the digit if one was pressed,
        or -1 on error or if the channel was disconnected.

        If musiconhold is playing before calling stream file
        it will be automatically stopped and will not be restarted after completion.

        Send the given file, allowing playback to be interrupted by the given
        digits, if any.  escape_digits is a string '12345' or a list  of
        ints [1,2,3,4,5] or strings ['1','2','3'] or mixed [1,'2',3,'4']
        If sample offset is provided then the audio will seek to sample
        offset before play starts.  Returns  digit if one was pressed
        or -1 if error/hangup channel.
        Remember, the file extension must not be included in the filename.
        """
        escape_digits = self._process_digit_list(escape_digits)
        message = await self.run_command(
            AGICommand.STREAM_FILE,
            filename,
            escape_digits,
            sample_offset
        )
        return message.result_to_char()

    async def tdd_mode(self, mode='off'):
        """
        Toggles TDD mode (for the deaf).

        Enable/Disable TDD transmission/reception on a channel.
        Returns 1 if successful, or 0 if channel is not TDD-capable.
        """
        message = await self.run_command(AGICommand.TDD_MODE, mode)
        if message.result == '0':
            raise AGIAppError(message='Channel %s is not TDD-capable')

    async def verbose(self, message, level=1):
        """
        Logs a message to the asterisk verbose log.

        Sends <message> to the console via verbose message system.
        <level> is the the verbose level (1-4).
        Always returns 1.
        """
        await self.run_command(AGICommand.VERBOSE, self._quote(message), level)

    async def wait_for_digit(self, timeout=DEFAULT_TIMEOUT):
        """
        Waits for a digit to be pressed.

        Waits for up to 'timeout' milliseconds for a channel to receive a DTMF digit.
        Waits up to timeout milliseconds for channel to receive a DTMF digit.
        Returns -1 on channel failure, 0 if no digit is received in the timeout,
        or the numerical value of the ascii of the digit if one is received.
        Use -1 for the timeout value if you desire the call to block indefinitely.
        """
        message = await self.run_command(AGICommand.WAIT_FOR_DIGIT, timeout)
        return message.result_to_char()



from aiohttp.base_protocol import BaseProtocol
from aiohttp.client_proto import ResponseHandler

from aioagi.log import agi_client_logger


class AGIResponseHandler(ResponseHandler):
    def __init__(self, *, loop=None):
        super().__init__(loop=loop)

        self._payload = None
        self._payload_parser = None

    def connection_lost(self, exc):
        if self._payload_parser is not None:
            try:
                self._payload_parser.feed_eof()
            except Exception:
                pass

        self.transport = self.writer = None
        self._should_close = True
        self._parser = None
        self._payload = None
        self._payload_parser = None
        self._reading_paused = False

        BaseProtocol.connection_lost(self, exc)

    def data_received(self, data):
        if not data:
            return

        if self._payload_parser is None:
            self._tail += data

        else:
            try:
                self._payload_parser.feed_data(data)
            except Exception as exc:
                agi_client_logger.exception('Command parser error.', exc_info=exc)
                self._should_close = True
                self.transport.close()
                self.set_exception(exc)
                return



from aiohttp.web_server import Server

from aioagi.protocol import AGIRequestHandler


class AGIServer(Server):
    def __call__(self):
        return AGIRequestHandler(self, loop=self._loop, **self._kwargs)



import functools
from aiohttp.helpers import AccessLogger


class AGIAccessLogger(AccessLogger):
    LOG_FORMAT = '%a - %t (%D) "%r" %s %b'

    def compile_format(self, log_format):
        """Translate log_format into form usable by modulo formatting

        All known atoms will be replaced with %s
        Also methods for formatting of those atoms will be added to
        _methods in apropriate order

        For example we have log_format = "%a %t"
        This format will be translated to "%s %s"
        Also contents of _methods will be
        [self._format_a, self._format_t]
        These method will be called and results will be passed
        to translated string format.

        Each _format_* method receive 'args' which is list of arguments
        given to self.log

        Exceptions are _format_e, _format_i and _format_o methods which
        also receive key name (by functools.partial)

        """
        # list of (key, method) tuples, we don't use an OrderedDict as users
        # can repeat the same key more than once
        methods = list()

        for atom in self.FORMAT_RE.findall(log_format):
            if atom[1] == '':
                format_key = self.LOG_FORMAT_MAP[atom[0]]
                m = getattr(self, '_format_%s' % atom[0])
            else:
                format_key = (self.LOG_FORMAT_MAP[atom[2]], atom[1])
                m = getattr(self, '_format_%s' % atom[2])
                m = functools.partial(m, atom[1])

            methods.append(self.KeyMethod(format_key, m))

        log_format = self.FORMAT_RE.sub(r'%s', log_format)
        log_format = self.CLEANUP_RE.sub(r'%\1', log_format)
        return log_format, methods

    @staticmethod
    def _format_r(request, response, time):
        return '{} {}'.format(request.method, request.path_qs)



import functools

from aioagi.client_proto import AGIResponseHandler
from aiohttp import TCPConnector


class AGITCPConnector(TCPConnector):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._factory = functools.partial(AGIResponseHandler, loop=self._loop)



from aiohttp import FlowControlDataQueue
from aiohttp.helpers import reify

from aioagi.agi_writer import AGIPayloadWriter

from aioagi.parser import AGIRequestMessage
from aioagi.protocol import AGIRequestHandler


class AGIRequest:
    def __init__(self,
                 message: AGIRequestMessage, protocol: AGIRequestHandler,
                 reader: FlowControlDataQueue, writer: AGIPayloadWriter):
        from aioagi.agi import AGI

        self.headers = self._headers = message.headers
        self.method = self._method = message.method
        self.rel_url = self._rel_url = message.url

        self.message = message
        self.protocol = protocol
        self.reader = reader
        self.writer = writer

        self._match_info = None
        self._cache = {}

        self.agi = AGI(self)

    @property
    def match_info(self):
        """Result of route resolving."""
        return self._match_info

    @property
    def transport(self):
        if self.protocol is None:
            return None
        return self.protocol.transport

    @reify
    def app(self):
        """Application instance."""
        return self._match_info.apps[-1]

    @reify
    def path(self):
        return self._rel_url.path

    @reify
    def path_qs(self):
        return str(self._rel_url)

    @reify
    def remote(self):
        """Remote IP of client initiated AGI request.

        The IP is resolved in this order:

        - overridden value by .clone(remote=new_remote) call.
        - peername of opened socket
        """
        if self.transport is None:
            return None
        peername = self.transport.get_extra_info('peername')
        if isinstance(peername, (list, tuple)):
            return peername[0]
        else:
            return peername





class AGIResponse:
    content = ''
    verbose = 6

    def __init__(self, status=200, body=b'', message=None, verbose=None):
        self._status = status
        self.body = body or self.content.encode()
        self.message = message
        self.verbose = verbose or self.verbose

    @property
    def status(self):
        return self._status

    @property
    def body(self):
        return self._body

    @body.setter
    def body(self, body):
        self._body = body

    @property
    def body_length(self):
        return len(self.body)



import asyncio

from aiohttp import DataQueue
from aiohttp.web_protocol import RequestHandler

from aioagi.log import agi_server_logger, agi_access_logger
from aioagi.parser import AGIRequestParser, AGIPayloadParser
from aioagi.helpers import AGIAccessLogger
from aioagi.response import AGIResponse
from aioagi.agi_writer import AGIPayloadWriter
from aioagi.exceptions import AGIException, AGIInternalServerError, AGIUnprocessableEntityError


class AGIRequestHandler(RequestHandler):
    def __init__(self, *args,
                 **kwargs):
        kwargs.update(
            logger=agi_server_logger,
            access_log_class=AGIAccessLogger,
            access_log=agi_access_logger,
            access_log_format=AGIAccessLogger.LOG_FORMAT,
        )
        super().__init__(*args, **kwargs)

        self._tcp_keepalive = False
        self._messages = asyncio.Queue(1)
        self._request_parser = AGIRequestParser(self, self._loop)
        self._stream_reader = None
        self._stream_writer = None

    def connection_made(self, transport):
        super(AGIRequestHandler, self).connection_made(transport)

        self._stream_reader = DataQueue(loop=self._loop)
        self._stream_writer = AGIPayloadWriter(self, loop=self._loop)

    def data_received(self, data):
        agi_server_logger.debug(data)
        if self._force_close or self._close:
            return

        # parse agi messages
        if self._payload_parser is None:
            try:
                message = self._request_parser.feed_data(data)
            except Exception as exc:
                self.log_exception('Request parser exception', exc_info=exc)
            else:
                if message:
                    self._payload_parser = AGIPayloadParser(self._stream_reader)
                    self._messages.put_nowait(message)

        # feed payload
        elif data:
            try:
                data = self._payload_parser.feed_data(data)
                eof, tail = data
                if eof:
                    self.force_close()

            except Exception as exc:
                self.log_exception('Payload parser exception', exc_info=exc)
                self.force_close()

    async def start(self):
        message = await self._messages.get()
        request = self._request_factory(message, self, self._stream_reader, self._stream_writer, self._task_handler)

        now = self._loop.time()
        try:
            response = await self._request_handler(request)

        except AGIInternalServerError as exc:
            response = exc
            self.logger.error('AGIException', exc_info=exc)

        except AGIException as exc:
            response = exc
            self.logger.debug('AGIException', exc_info=exc)

        except asyncio.CancelledError:
            self.logger.debug('Ignored premature client disconnection')
            self.log_access(request, AGIUnprocessableEntityError(), self._loop.time() - now)
            if self.transport:
                self.transport.write(b'HANGUP\n')
            return

        except Exception as exc:
            self.logger.error('Unhandled exception', exc_info=exc)
            response = AGIInternalServerError()

        if not response:
            response = AGIResponse()

        if self.transport and response.message:
            try:
                self.transport.write('VERBOSE "{message}" {verbose}\n'.format(
                    message=response.message,
                    verbose=response.verbose,
                ).encode())

            except Exception as exc:
                self.logger.warning('Message send error', exc_info=exc)

        if self.transport and response.body:
            self.transport.write(response.body)

        self.log_access(request, response, self._loop.time() - now)
        if not self._force_close:
            if self.transport is not None:
                self.transport.close()



import re
import yarl
import collections
from enum import IntEnum

from aioagi.log import agi_server_logger, agi_client_logger
from aioagi.commands import AGICommand
from aioagi.exceptions import AGIAppError

AGIRequestMessage = collections.namedtuple(
    'AGIRequestMessage',
    ['method', 'path', 'headers', 'url']
)


class AGICode(IntEnum):
    TRYING = 100

    OK = 200

    MEMORY_ALLOCATION_FAILURE = 503
    INVALID_COMMAND = 510
    COMMAND_NOT_PERMITTED = 511
    INVALID_COMMAND_SYNTAX = 520

    # Internal codes
    UNKNOWN = 1000


class AGIMessage:
    def __init__(self, status: AGICode, result: str, data: dict, info: str=''):
        self.status = status
        self.result = result
        self.data = data
        self.info = info

    def __repr__(self):
        return '{class_name}(status={status}, result={result}, data={data}, info="{info}")'.format(
            class_name=self.__class__.__name__,
            status=self.status,
            result=self.result,
            data=self.data,
            info=self.info,
        )

    def result_as_char(self):
        if self.result == '0':
            return ''

        try:
            return chr(int(self.result))
        except (TypeError, ValueError):
            raise AGIAppError(message='Unable to convert result to char: {}'.format(self.result))

    def result_as_int(self):
        try:
            return int(self.result)
        except (TypeError, ValueError):
            raise AGIAppError(message='Unable to convert result to int: {}'.format(self.result))

    def result_as_bool(self):
        try:
            return bool(int(self.result))
        except (TypeError, ValueError):
            raise AGIAppError(message='Unable to convert result to int: {}'.format(self.result))

    def result_to_char(self):
        self.result = self.result_as_char()
        return self

    def result_to_int(self):
        self.result = self.result_as_int()
        return self

    def result_to_bool(self):
        self.result = self.result_as_bool()
        return self

    def as_data(self):
        result = ' result={}'.format(self.result) if self.result is not None else ''
        data = ' ({data})'.format(data=self.data.pop('data')) if 'data' in self.data else ''

        key_value = ['{}={}'.format(key, value) for key, value in self.data.items() if key != 'data']
        key_value = ' {}'.format(key_value) if key_value else ''

        return '{status}{result}{data}{key_value}\n'.format(
            status=self.status,
            result=result,
            data=data,
            key_value=key_value,
            info=' {}'.format(self.info) if self.info else ''
        )


class AGIClientMessage:
    def __init__(self, command: AGICommand, args):
        self.command = command
        self.args = args

    def __repr__(self):
        return '{class_name}(command={command}, args={args})'.format(
            class_name=self.__class__.__name__,
            command=self.command,
            args=self.args,
        )


class AGIRequestParser:
    method_p = re.compile('[^a-zA-Z0-9_]')

    def __init__(self, protocol, loop):
        self.protocol = protocol
        self.loop = loop

        self._tail = b''
        self._lines = []

    def feed_data(self, data, sep=b'\n', empty=b''):
        message = None
        if self._tail:
            data, self._tail = self._tail + data, empty

        data_len = len(data)
        start_pos = 0

        while start_pos < data_len:
            pos = data.find(sep, start_pos)
            if pos >= start_pos:
                # line found
                self._lines.append(data[start_pos:pos])
                start_pos = pos + 1
                if self._lines[-1] == empty:
                    try:
                        message = self.parse_message(self._lines[:-1])
                    finally:
                        self._lines.clear()

                    break

            else:
                self._tail = data[start_pos:]
                break

        return message

    def parse_message(self, lines):
        """
        Raw protocol:
            agi_network: yes
            agi_network_script: agi/
            agi_request: agi://10.99.99.55:8080/agi/
            agi_channel: SIP/95000-00000005
            agi_language: ru
            agi_type: SIP
            agi_uniqueid: 1502266948.5
            agi_version: 11.14.2
            agi_callerid: 95000
            agi_calleridname: test
            agi_callingpres: 0
            agi_callingani2: 0
            agi_callington: 0
            agi_callingtns: 0
            agi_dnid: 770
            agi_rdnis: unknown
            agi_context: from-internal
            agi_extension: 770
            agi_priority: 2
            agi_enhanced: 0.0
            agi_accountcode:
            agi_threadid: 139689736754944

        headers = {
            'agi_network': 'yes',
            'agi_network_script': 'agi/',
            'agi_request': 'agi://10.99.99.55:8080/agi/',
            'agi_channel': 'SIP/95000-00000005',
            'agi_language': 'ru',
            'agi_type': 'SIP',
            'agi_uniqueid': '1502266948.5',
            'agi_version': '11.14.2',
            'agi_callerid': '95000',
            'agi_calleridname': 'test',
            'agi_callingpres': '0',
            'agi_callingani2': '0',
            'agi_callington': '0',
            'agi_callingtns': '0',
            'agi_dnid': '770',
            'agi_rdnis': 'unknown',
            'agi_context': 'from-internal',
            'agi_extension': '770',
            'agi_priority': '2',
            'agi_enhanced': '0.0',
            'agi_accountcode': '',
            'agi_threadid': '139689736754944',
        }
        """
        headers = collections.OrderedDict([line.decode().split(': ', 1) for line in lines])
        path = headers.get('agi_request', '')

        method = headers.get('agi_type', 'GET').replace(' ', '_').replace('-', '_')
        method = self.method_p.sub('', method).upper()

        return AGIRequestMessage(method, path, headers, yarl.URL(path))


class AGIPayloadParser:
    """
    Parse AGI results.
    """
    status_p = re.compile(r'(?P<status>^\d*)(\s|-)*(?P<response>.*)', re.DOTALL)
    kwargs_p = re.compile(r'(?P<key>\w+)=(?P<value>[^\s]*)')
    data_p = re.compile(r'result=[^\s]+\s*\((?P<data>.*?)\)')

    def __init__(self, payload):
        self.payload = payload
        self._tail = b''

    def feed_eof(self):
        self.payload.feed_eof()

    def feed_data(self, data, sep=b'\n'):
        if self._tail:
            data, self._tail = self._tail + data, b''

        start_pos = 0
        data_len = len(data)

        lines = []
        while start_pos < data_len:
            pos = data.find(sep, start_pos)
            if pos >= start_pos:
                line = data[start_pos:pos]
                lines.append(line)
                start_pos = pos + 1

            else:
                self._tail = data[start_pos:]
                break

        if not lines:
            return False, self._tail

        eof, message, tail = self.parse_message(lines)

        if eof:
            self.payload.feed_eof()

        elif message:
            self.payload.feed_data(message, len(data[:start_pos]))

        elif tail:
            self._tail = tail

        return eof, self._tail

    def parse_message(self, lines):
        """
        AGI Result examples::

            100 result=0 Trying...
            ?
            200 result=0 Gosub complete
            ?
            200 result=%d Gosub failed

            200 result=-1 Gosub label not found"

            200 result=0

            200 result=0 endpos=24807

            200 result=49 endpos=8640

            200 result=-1

            200 result=132456

            200 result=1 (testvariable)

            200 result= (timeout)

            503 result=-2 Memory allocation failure

            510 Invalid or unknown command

            511 Command Not Permitted on a dead channel or intercept routine

            # GET DATA
            520-Invalid command syntax. Proper usage follows:
            Stream the given <replaceable>file</replaceable>, and receive DTMF data.
            Returns the digits received from the channel at the other end.520 End of proper usage.

            HANGUP

        From source code example:
            'ast_agi_send(agi->fd, chan, "%d %s\n", resultcode, ami_res);',
            'ast_agi_send(agi->fd, chan, "%s\n", c->usage);',
            'ast_agi_send(agi->fd, chan, "100 result=0 Trying...\n");',
            'ast_agi_send(agi->fd, chan, "200 result=%c\n", num_deleted > 0 ? \'0\' : \'1\');',
            'ast_agi_send(agi->fd, chan, "200 result=%c\n", res ? \'0\' : \'1\');',
            'ast_agi_send(agi->fd, chan, "200 result=%d (dtmf) endpos=%ld\n", f->subclass.integer, sample_offset);',
            'ast_agi_send(agi->fd, chan, "200 result=%d (hangup) endpos=%ld\n", -1, sample_offset);',
            'ast_agi_send(agi->fd, chan, "200 result=%d (hangup)\n", res);',
            'ast_agi_send(agi->fd, chan, "200 result=%d (randomerror) endpos=%ld\n", res, sample_offset);',
            'ast_agi_send(agi->fd, chan, "200 result=%d (timeout) endpos=%ld\n", res, sample_offset);',
            'ast_agi_send(agi->fd, chan, "200 result=%d (timeout)\n", res);',
            'ast_agi_send(agi->fd, chan, "200 result=%d (waitfor) endpos=%ld\n", res,sample_offset);',
            'ast_agi_send(agi->fd, chan, "200 result=%d (writefile)\n", res);',
            'ast_agi_send(agi->fd, chan, "200 result=%d Gosub failed\n", res);',
            'ast_agi_send(agi->fd, chan, "200 result=%d endpos=%ld\n", res, offsetms);',
            'ast_agi_send(agi->fd, chan, "200 result=%d endpos=%ld\n", res, sample_offset);',
            'ast_agi_send(agi->fd, chan, "200 result=%d\n", res);',
            'ast_agi_send(agi->fd, chan, "200 result=%s (timeout)\n", data);',
            'ast_agi_send(agi->fd, chan, "200 result=%s\n", data);',
            'ast_agi_send(agi->fd, chan, "200 result=%u\n", ast_channel_state(chan));',
            'ast_agi_send(agi->fd, chan, "200 result=%u\n", snapshot->state);',
            'ast_agi_send(agi->fd, chan, "200 result=-1 Gosub label not found\n");',
            'ast_agi_send(agi->fd, chan, "200 result=-1 endpos=%ld\n", sample_offset);',
            'ast_agi_send(agi->fd, chan, "200 result=-1\n");',
            'ast_agi_send(agi->fd, chan, "200 result=0 endpos=%ld\n", current_offset);',
            'ast_agi_send(agi->fd, chan, "200 result=0\n");',
            'ast_agi_send(agi->fd, chan, "200 result=1 (%s) endpos=%ld\n", reason, current_offset);',
            'ast_agi_send(agi->fd, chan, "200 result=1 (%s)\n", ast_str_buffer(buf));',
            'ast_agi_send(agi->fd, chan, "200 result=1 (%s)\n", ast_str_buffer(str));',
            'ast_agi_send(agi->fd, chan, "200 result=1 (%s)\n", buf);',
            'ast_agi_send(agi->fd, chan, "200 result=1 (%s)\n", ret);',
            'ast_agi_send(agi->fd, chan, "200 result=1 (digit) digit=%c endpos=%ld\n", dtmf, current_offset);',
            'ast_agi_send(agi->fd, chan, "200 result=1 (speech) endpos=%ld results=%d %s\n", current_offset, i, tmp);',
            'ast_agi_send(agi->fd, chan, "200 result=1\n");',
            'ast_agi_send(agi->fd, chan, "503 result=-2 Memory allocation failure\n");',
            'ast_agi_send(agi->fd, chan, "520 End of proper usage.\n");',
            'ast_agi_send(agi->fd, chan, "520 Invalid command syntax.  Proper usage not available.\n");',
            'ast_agi_send(agi->fd, chan, "520-Invalid command syntax.  Proper usage follows:\n");',
            'ast_agi_send(agi->fd, chan, "HANGUP\n");',
            'ast_agi_send(fd, chan, "\n");',
            'ast_agi_send(fd, chan,
                "agi_accountcode: %s\n", ast_channel_accountcode(chan) ? ast_channel_accountcode(chan) : "");',
            'ast_agi_send(fd, chan, "agi_arg_%d: %s\n", count, argv[count]);',
            'ast_agi_send(fd, chan, "agi_callingani2: %d\n", ast_channel_caller(chan)->ani2);',
            'ast_agi_send(fd, chan, "agi_callingtns: %d\n", ast_channel_dialed(chan)->transit_network_select);',
            'ast_agi_send(fd, chan, "agi_callington: %d\n", ast_channel_caller(chan)->id.number.plan);',
            'ast_agi_send(fd, chan, "agi_channel: %s\n", ast_channel_name(chan));',
            'ast_agi_send(fd, chan, "agi_context: %s\n", ast_channel_context(chan));',
            'ast_agi_send(fd, chan, "agi_dnid: %s\n", S_OR(ast_channel_dialed(chan)->number.str, "unknown"));',
            'ast_agi_send(fd, chan, "agi_enhanced: %s\n", enhanced ? "1.0" : "0.0");',
            'ast_agi_send(fd, chan, "agi_extension: %s\n", ast_channel_exten(chan));',
            'ast_agi_send(fd, chan, "agi_language: %s\n", ast_channel_language(chan));',
            'ast_agi_send(fd, chan, "agi_priority: %d\n", ast_channel_priority(chan));',
            'ast_agi_send(fd, chan, "agi_request: %s\n", request);',
            'ast_agi_send(fd, chan, "agi_threadid: %ld\n", (long)pthread_self());',
            'ast_agi_send(fd, chan, "agi_type: %s\n", ast_channel_tech(chan)->type);',
            'ast_agi_send(fd, chan, "agi_uniqueid: %s\n", ast_channel_uniqueid(chan));',
            'ast_agi_send(fd, chan, "agi_version: %s\n", ast_get_version());',
            'ast_agi_send(s, NULL, "agi_network_script: %s\n", script);',
            'ast_agi_send;'
        """
        message = b'\n'.join(lines).decode()
        if AGICommand.HANGUP in message:
            return True, None, None

        result = None
        match = self.status_p.search(message)
        if match:
            data = match.groupdict()
            message_data = {}
            status_int = int(data['status'])
            try:
                status = AGICode(status_int)
            except ValueError:
                status = AGICode.UNKNOWN
                message_data['_status'] = status_int
                message_data['_message'] = message
                agi_server_logger.warning('Unhandled code or undefined response: %s', status)

            info = self.kwargs_p.sub('', self.data_p.sub('', data['response'])).strip()

            if status == AGICode.OK:
                for key, value in self.kwargs_p.findall(data['response']):
                    if key == 'result':
                        result = value
                    else:
                        message_data[key] = value

                data_match = self.data_p.search(data['response'])
                if data_match:
                    message_data['data'] = data_match.group(1)

            elif status == AGICode.INVALID_COMMAND_SYNTAX:
                # AGI Usage error
                if (data['response'].endswith('Proper usage not available.') or
                        data['response'].endswith('520 End of proper usage.')):
                    pass

                else:
                    return False, None, message.encode()

            elif status in (AGICode.TRYING, AGICode.INVALID_COMMAND, AGICode.COMMAND_NOT_PERMITTED):
                # Without result and data.
                pass

            elif status == AGICode.MEMORY_ALLOCATION_FAILURE:
                agi_server_logger.critical('Unhandled code or undefined response: %s', status)

        else:
            agi_server_logger.warning('Undefined response: %s', message)
            return True, None, None

        return False, AGIMessage(status, result, message_data, info), None


class AGIClientPayloadParser:
    def __init__(self, payload):
        self.payload = payload
        self._tail = b''

    def feed_eof(self):
        self.payload.feed_eof()

    def feed_data(self, data, sep=b'\n'):
        if self._tail:
            data, self._tail = self._tail + data, b''

        start_pos = 0
        data_len = len(data)

        lines = []
        while start_pos < data_len:
            pos = data.find(sep, start_pos)
            if pos >= start_pos:
                line = data[start_pos:pos]
                lines.append(line)
                start_pos = pos + 1

            else:
                self._tail = data[start_pos:]
                break

        if not lines:
            return False, self._tail

        eof = False

        for eof, message, payload in self.parse_message(lines):
            if eof:
                self.payload.feed_eof()
                break
            elif message:
                self.payload.feed_data(message, len(payload))

        return eof, self._tail

    def parse_message(self, lines):
        for line in lines:
            eof = False
            data = line.decode('utf-8')

            if line.lower().startswith(b'failure'):
                yield True, None, ''

            for command in AGICommand.__members__.values():
                if not data.startswith(command.value):
                    continue

                if command == AGICommand.HANGUP:
                    eof = True

                args = data.split(command.value)[-1].strip().split()
                yield eof, AGIClientMessage(command, args), line
                break

            else:
                agi_client_logger.warning('Invalid command: {}'.format(data))



import sys
import attr
import asyncio
import traceback

from multidict import MultiDict

from yarl import URL
from aiohttp.helpers import CeilTimeout
from aiohttp.streams import FlowControlDataQueue, EofStream
from aiohttp.connector import Connection
from aiohttp.client_reqrep import RequestInfo

from aioagi.log import agi_client_logger
from aioagi.hdrs import AGIMethod
from aioagi.parser import AGIClientPayloadParser, AGIMessage
from aioagi.agi_writer import AGIPayloadWriter
from aioagi.client_exceptions import AGIServerTimeoutError


@attr.s(slots=True, frozen=True)
class ConnectionKey:
    # the key should contain an information about used proxy / TLS
    # to prevent reusing wrong connections from a pool
    host = attr.ib(type=str)
    port = attr.ib(type=int)
    is_ssl = attr.ib(type=bool)
    ssl = attr.ib()  # SSLContext or None
    proxy = attr.ib()  # URL or None
    proxy_auth = attr.ib()  # BasicAuth
    proxy_headers_hash = attr.ib(type=int)  # hash(CIMultiDict)


class AGIClientRequest:
    def __init__(self, method: AGIMethod, url: URL, *,
                 params=None, headers=None,
                 timeout=None, response_class=None,
                 session=None, loop=None):

        if loop is None:
            loop = asyncio.get_event_loop()

        assert isinstance(url, URL), url
        self._session = session
        self._connector = session.connector

        if params:
            q = MultiDict(url.query)
            url2 = url.with_query(params)
            q.extend(url2.query)
            url = url.with_query(q)
        self.url = url.with_fragment(None)

        self.original_url = url
        self.method = method
        self.loop = loop
        self.length = None
        self.timeout = timeout
        self.response_class = response_class or AGIClientResponse

        if loop.get_debug():
            self._source_traceback = traceback.extract_stack(sys._getframe(1))

        self.ssl = False
        self.proxy = False

        self.headers = headers or {}
        self.set_headers()

        self._protocol = None
        self._connection = None
        self._writer = None
        self._reader = None
        self._response = None

    async def __aenter__(self):
        return await self.start()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._response.close()

    @property
    def connection_key(self):
        return ConnectionKey(self.host, self.port, self.ssl, self.ssl, None, None, None)

    @property
    def host(self):
        return self.url.host

    @property
    def port(self):
        return self.url.port

    @property
    def request_info(self):
        return RequestInfo(self.url, self.method, self.headers)

    def is_ssl(self):
        return False

    def set_headers(self):
        """
        headers = {
            'agi_network': 'yes',
            'agi_network_script': 'agi/test/?a=asdf&b=jkl',
            'agi_request': 'agi://10.197.101.150:8082/agi/test/?a=asdf&b=jkl',
            'agi_channel': 'SIP/95000-00000005',
            'agi_language': 'ru',
            'agi_type': 'SIP',
            'agi_uniqueid': '1502266948.5',
            'agi_version': '11.14.2',
            'agi_callerid': '95000',
            'agi_calleridname': 'test',
            'agi_callingpres': '0',
            'agi_callingani2': '0',
            'agi_callington': '0',
            'agi_callingtns': '0',
            'agi_dnid': '770',
            'agi_rdnis': 'unknown',
            'agi_context': 'from-internal',
            'agi_extension': '770',
            'agi_priority': '2',
            'agi_enhanced': '0.0',
            'agi_accountcode': '',
            'agi_threadid': '139689736754944',
        }
        """
        self.headers['agi_type'] = self.method
        self.headers['agi_request'] = str(self.url)
        self.headers['agi_network'] = 'yes'
        self.headers['agi_network_script'] = self.url.path_qs[1:]

    async def start(self):
        # connection timeout
        try:
            with CeilTimeout(self.timeout.connect, loop=self.loop):
                connection = await self._connector.connect(self, None, self.timeout)

        except asyncio.TimeoutError as exc:
            raise AGIServerTimeoutError(
                'Connection timeout '
                'to host {0}'.format(self.url)) from exc
        else:
            self._protocol = connection.protocol
            self._connection = connection

        # connection.writer.set_tcp_nodelay(True)
        try:
            self._response = await self.send(connection)
        except Exception as exc:
            agi_client_logger.error('Request error:', exc_info=exc)
            await connection.close()
            raise

        return self._response

    async def send(self, connection: Connection):
        self._protocol = connection.protocol
        self._connection = connection

        self._writer = AGIPayloadWriter(connection.protocol, loop=self.loop)
        self._reader = FlowControlDataQueue(connection.protocol, loop=self.loop)

        connection.protocol.set_parser(AGIClientPayloadParser(self._reader), self._reader)

        # headers
        await self._writer.write_headers(self.headers)
        await self._writer.drain()

        self.response = self.response_class(
            self.method, self.original_url, connection,
            writer=self._writer, reader=self._reader,
            request_info=self.request_info,
            loop=self.loop,
        )

        return self.response


class AGIClientResponse:
    def __init__(self, method: AGIMethod, url: URL, connection: Connection,
                 writer: AGIPayloadWriter, reader: FlowControlDataQueue,
                 request_info, loop):

        self._method = method
        self._url = url
        self._connection = connection
        self._protocol = connection.protocol
        self._closed = False
        self._writer = writer
        self._reader = reader
        self._request_info = request_info
        self._waiting = None
        self._loop = loop
        self._exception = None

    def __aiter__(self):
        return self

    async def __anext__(self):
        msg = await self.receive()
        if not msg:
            raise StopAsyncIteration()

        return msg

    @property
    def closed(self):
        return self._closed

    async def close(self):
        if self._waiting is not None and not self._closed:
            self._reader.feed_eof()
            await self._waiting

        if not self._closed:
            self._closed = True
            if self._loop is None or self._loop.is_closed():
                return

            if self._connection is not None:
                self._connection.close()
                self._connection = None

        else:
            return False

    async def send(self, message: AGIMessage):
        await self._writer.write(message.as_data())

    async def send_str(self, data):
        await self._writer.write(data)

    async def receive(self):
        if self._closed:
            return None

        self._waiting = self._loop.create_future()
        try:
            msg = await self._reader.read()
        except EofStream:
            return None
        except (asyncio.CancelledError, asyncio.TimeoutError):
            raise
        except Exception as exc:
            self._exception = exc
            agi_client_logger.exception('Message read error.', exc_info=exc)
            raise

        finally:
            waiter = self._waiting
            self._waiting = None
            waiter.set_result(True)

        return msg



import enum


@enum.unique
class AGICommand(str, enum.Enum):
    ANSWER = 'ANSWER'
    CHANNEL_STATUS = 'CHANNEL STATUS'
    CONTROL_STREAM_FILE = 'CONTROL STREAM FILE'
    DATABASE_GET = 'DATABASE GET'
    DATABASE_PUT = 'DATABASE PUT'
    DATABASE_DEL = 'DATABASE DEL'
    DATABASE_DELTREE = 'DATABASE DELTREE'
    EXEC = 'EXEC'
    GET_DATA = 'GET DATA'
    GET_FULL_VARIABLE = 'GET FULL VARIABLE'
    GET_OPTION = 'GET OPTION'
    GET_VARIABLE = 'GET VARIABLE'
    GOSUB = 'GOSUB'
    HANGUP = 'HANGUP'
    NOOP = 'NOOP'
    RECEIVE_CHAR = 'RECEIVE CHAR'
    RECEIVE_TEXT = 'RECEIVE TEXT'
    RECORD_FILE = 'RECORD FILE'
    SAY_ALPHA = 'SAY ALPHA'
    SAY_DATE = 'SAY DATE'
    SAY_DATETIME = 'SAY DATETIME'
    SAY_DIGITS = 'SAY DIGITS'
    SAY_NUMBER = 'SAY NUMBER'
    SAY_PHONETIC = 'SAY PHONETIC'
    SAY_TIME = 'SAY TIME'
    SEND_IMAGE = 'SEND IMAGE'
    SEND_TEXT = 'SEND TEXT'
    SET_AUTOHANGUP = 'SET AUTOHANGUP'
    SET_CALLERID = 'SET CALLERID'
    SET_CONTEXT = 'SET CONTEXT'
    SET_MUSIC = 'SET MUSIC'
    SET_EXTENSION = 'SET EXTENSION'
    SET_PRIORITY = 'SET PRIORITY'
    SET_VARIABLE = 'SET VARIABLE'
    STREAM_FILE = 'STREAM FILE'
    TDD_MODE = 'TDD MODE'
    VERBOSE = 'VERBOSE'
    WAIT_FOR_DIGIT = 'WAIT FOR DIGIT'



from aioagi.response import AGIResponse


class AGIException(AGIResponse, Exception):
    pass


class AGIHangup(AGIException):
    status = 204


class AGIResultHangup(AGIHangup):
    pass


class AGIConnectHangup(AGIHangup):
    pass


class AGIMove(AGIException):
    verbose = 3
    status = 300

    def __init__(self, url=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if url:
            self.message = 'Location: {}'.format(url)


class FailureAGIError(AGIException):
    content = 'FAILURE\n'
    verbose = 3


# ############
# Client Error
# ############


class AGIBadRequest(FailureAGIError):
    status = 400


class AGINotAuthenticated(FailureAGIError):
    status = 401


class AGIPermissionDenied(FailureAGIError):
    status = 403


# Resolver exceptions
class AGIMethodNotAllowed(FailureAGIError):
    status = 405


class AGINotFound(FailureAGIError):
    status = 404


class AGIAppError(FailureAGIError):
    status = 421


class AGIUnprocessableEntityError(FailureAGIError):
    status = 422


# #############
# Server errors
# #############
class AGIInternalServerError(FailureAGIError):
    status = 500
    verbose = 2

    def __init__(self, error='', *args, **kwargs):
        self.error = error
        super().__init__(*args, **kwargs)


# Command errors
class AGIMemoryAllocationFailure(AGIInternalServerError):
    status = 503


class AGIInvalidCommand(AGIInternalServerError):
    status = 510


class AGICommandNotPermittedError(AGIInternalServerError):
    status = 511


class AGIUsageError(AGIInternalServerError):
    status = 520



import asyncio
import warnings

from yarl import URL

from aiohttp import ClientTimeout
from aiohttp.helpers import sentinel
from aiohttp.client import DEFAULT_TIMEOUT

from aioagi.hdrs import AGIMethod
from aioagi.connector import AGITCPConnector
from aioagi.client_exceptions import AGIInvalidURL, AGIInvalidMethod
from aioagi.client_reqrep import AGIClientRequest, AGIClientResponse


class AGIClientSession:
    def __init__(self, *, connector=None, loop=None,
                 headers=None,
                 request_class=AGIClientRequest, response_class=AGIClientResponse,
                 connector_owner=True, raise_for_status=False,
                 timeout=None):

        if loop is None:
            loop = asyncio.get_event_loop()

        self._loop = loop

        if connector is None:
            connector = AGITCPConnector(loop=self._loop)

        self._connector = connector
        self._connector_owner = connector_owner

        self._timeout = timeout or DEFAULT_TIMEOUT
        self._raise_for_status = raise_for_status

        self._default_headers = headers or {}

        self._request_class = request_class
        self._response_class = response_class

    def __del__(self, _warnings=warnings):
        if not self.closed:
            self._connector.close()

            _warnings.warn("Unclosed client session {!r}".format(self),
                           ResourceWarning)
            context = {'client_session': self,
                       'message': 'Unclosed client session'}

            self._loop.call_exception_handler(context)

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    def request(self, method, url, *, params=None, headers=None, timeout=sentinel):
        if not isinstance(method, AGIMethod):
            try:
                method = AGIMethod(method)
            except ValueError:
                raise AGIInvalidMethod(method)

        try:
            url = URL(url)
        except ValueError:
            raise AGIInvalidURL(url)

        url = url.with_fragment(None)

        request_headers = self._default_headers.copy()
        headers and request_headers.update(headers)

        if timeout is sentinel:
            timeout = self._timeout
        else:
            if not isinstance(timeout, ClientTimeout):
                timeout = ClientTimeout(total=timeout)

        return self._request_class(
            method, url,
            params=params, headers=request_headers,
            timeout=timeout, response_class=self._response_class,
            session=self, loop=self._loop)

    def local(self, url, **kwargs):
        return self.request(AGIMethod.LOCAL, url, **kwargs)

    def sip(self, url, **kwargs):
        return self.request(AGIMethod.SIP, url, **kwargs)

    def dahdi(self, url, **kwargs):
        return self.request(AGIMethod.DAHDI, url, **kwargs)

    def iax2(self, url, **kwargs):
        return self.request(AGIMethod.IAX2, url, **kwargs)

    def motif(self, url, **kwargs):
        return self.request(AGIMethod.MOTIF, url, **kwargs)

    def misdn(self, url, **kwargs):
        return self.request(AGIMethod.MISDN, url, **kwargs)

    def ustm(self, url, **kwargs):
        return self.request(AGIMethod.USTM, url, **kwargs)

    def skinny(self, url, **kwargs):
        return self.request(AGIMethod.SKINNY, url, **kwargs)

    def h323(self, url, **kwargs):
        return self.request(AGIMethod.H323, url, **kwargs)

    def ooh323(self, url, **kwargs):
        return self.request(AGIMethod.OOH323, url, **kwargs)

    def gtalk(self, url, **kwargs):
        return self.request(AGIMethod.GTALK, url, **kwargs)

    def jingle(self, url, **kwargs):
        return self.request(AGIMethod.JINGLE, url, **kwargs)

    @property
    def loop(self):
        """Session's loop."""
        return self._loop

    @property
    def closed(self):
        """Is client session closed.

        A readonly property.
        """
        return self._connector is None or self._connector.closed

    @property
    def connector(self):
        """Connector instance used for the session."""
        return self._connector

    def detach(self):
        """Detach connector from session without closing the former.

        Session is switched to closed state anyway.
        """
        self._connector = None

    async def close(self):
        """Close underlying connector.

        Release all acquired resources.
        """
        if not self.closed:
            if self._connector_owner:
                self._connector.close()
            self._connector = None



import logging

agi_access_logger = logging.getLogger('aioagi.access')
agi_client_logger = logging.getLogger('aioagi.client')
agi_internal_logger = logging.getLogger('aioagi.internal')
agi_server_logger = logging.getLogger('aioagi.server')
agi_app_logger = logging.getLogger('aioagi.app')




import aioagi

from setuptools import setup, find_packages

README = open('README.rst').read()

tests_require = ['coverage', 'pytest', 'pytest-asyncio', 'pytest-sugar', 'pytest-cov']


setup(
    name='aioagi',
    version=aioagi.VERSION,
    description='Async agi client/server framework (asyncio)',
    long_description=README,
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Operating System :: POSIX',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Topic :: Communications :: Internet Phone',
        'Topic :: Communications :: Telephony',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Framework :: AsyncIO',
    ],
    author='Shakurov Vadim Vladimirovich',
    author_email='apelsinsd@gmail.com',
    url='https://gitlab.com/VadimShakurov/aioagi.git',
    license='Apache 2',
    keywords='aiogi asyncio asterisk telephony voip',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'aiohttp>=3.3.0',
    ],
    tests_require=tests_require,
    extras_require={
        'dev': [
            'ipdb',
            'ipython',
        ],
        'testing': tests_require
    },
)



from aiohttp.web_urldispatcher import UrlDispatcher, MatchInfoError

from aiohttp import web

from aioagi.hdrs import AGIMethod
from aioagi.exceptions import AGIMethodNotAllowed, AGINotFound
from aioagi.request import AGIRequest


class AGIUrlDispatcher(UrlDispatcher):
    async def resolve(self, request: AGIRequest):
        allowed_methods = set()

        for resource in self._resources:
            match_dict, allowed = await resource.resolve(request)
            if match_dict is not None:
                return match_dict
            else:
                allowed_methods |= allowed
        else:
            if allowed_methods:
                return MatchInfoError(AGIMethodNotAllowed(
                    message='{}: {}'.format(request.method, allowed_methods)))
            else:
                return MatchInfoError(AGINotFound())


class AGIView(web.View):
    def __init__(self, request: AGIRequest):
        self.agi = request.agi
        super().__init__(request)

    async def _iter(self):
        if self.request.method not in AGIMethod.all():
            self._raise_allowed_methods()

        method = getattr(self, self.request.method.lower(), None)
        if method is None:
            self._raise_allowed_methods()
        return await method()

    def _raise_allowed_methods(self):
        allowed_methods = {m for m in AGIMethod.all() if hasattr(self, m.lower())}
        raise AGIMethodNotAllowed(message='{}: {}'.format(self.request.method, allowed_methods))



from functools import partial

from aiohttp.web import Application
from aiohttp.abc import AbstractAccessLogger

from aioagi.log import agi_app_logger
from aioagi.server import AGIServer
from aioagi.request import AGIRequest
from aioagi.helpers import AGIAccessLogger
from aioagi.urldispathcer import AGIUrlDispatcher


class AGIApplication(Application):
    server_cls = AGIServer

    def __init__(self, logger=agi_app_logger, **kwargs):
        kwargs.update(router=AGIUrlDispatcher())
        super().__init__(logger=logger, **kwargs)

    def _make_request(self, message, payload, protocol, writer, task,
                      _cls=AGIRequest):
        return _cls(message, payload, protocol, writer)

    def _make_handler(self,
                      loop=None,
                      access_log_class=AGIAccessLogger,
                      **kwargs):

        if not issubclass(access_log_class, AbstractAccessLogger):
            raise TypeError(
                'access_log_class must be subclass of '
                'aiohttp.abc.AbstractAccessLogger, got {}'.format(
                    access_log_class))

        self._set_loop(loop)
        self.freeze()

        kwargs['debug'] = self.debug
        if self._handler_args:
            for k, v in self._handler_args.items():
                kwargs[k] = v

        return self.server_cls(self._handle, request_factory=self._make_request,
                               access_log_class=access_log_class,
                               loop=self.loop, **kwargs)

    async def _handle(self, request):
        match_info = await self._router.resolve(request)
        match_info.add_app(self)
        match_info.freeze()

        request._match_info = match_info
        handler = match_info.handler

        if self._run_middlewares:
            for app in match_info.apps[::-1]:
                for m, new_style in app._middlewares_handlers:
                    if new_style:
                        handler = partial(m, handler=handler)
                    else:
                        handler = await m(app, handler)

        resp = await handler(request)

        return resp
